﻿using System.Collections.Generic;
using System.Linq;

namespace PaintShopSchedulingCarrierChangeOptimizer
{
   /// <summary>
    /// This provides a common interface for various implementations 
    /// which optimizes the carrier changes.
    /// It also contains some common used methods.
    /// </summary>
    internal abstract class AbstractCarrierChangeOptimizer : ICarrierChangeOptimizer
    {

        /// <summary>
        /// This function optimizes the carrier changes for two given rounds <paramref name="currentRound"/> and <paramref name="nextRound"/>.
        /// It optimize the number of carrier insertions and removals between the given round and its successor.
        /// It defines which carriers should be removed/inserted directly 
        /// in the given lists <paramref name="currentRound"/> respectively <paramref name="nextRound"/>.
        /// </summary>
        /// <param name="currentRound"></param>
        /// <param name="nextRound"></param>
        public void Solve(IRoundInformation currentRound,
            IRoundInformation nextRound)
        {
            IList<ICarrierInsertionRemovalFlagInformation> preparedCurrentRoundCarrierList = currentRound.CarrierInsertionRemovalFlags;
            IList<ICarrierInsertionRemovalFlagInformation> preparedNextRoundCarrierList = nextRound.CarrierInsertionRemovalFlags;

            Optimize(preparedCurrentRoundCarrierList, preparedNextRoundCarrierList);
        }

        /// <summary>
        /// Internal method which does the actual optimization.
        /// </summary>
        /// <param name="preparedCurrentRoundCarrierList"></param>
        /// <param name="preparedNextRoundCarrierList"></param>
        protected abstract void Optimize(IList<ICarrierInsertionRemovalFlagInformation> preparedCurrentRoundCarrierList,
            IList<ICarrierInsertionRemovalFlagInformation> preparedNextRoundCarrierList);

        /// <summary>
        /// Calculates the number of optimzed carrier changes for for two given rounds <paramref name="currentRound"/> and <paramref name="nextRound"/>.
        /// </summary>
        /// <param name="currentRound"></param>
        /// <param name="nextRound"></param>
        /// <returns>An object which contains the number of determined carrier removals and carrier insertions.</returns>
        public abstract INumberOfCarrierChanges EvaluateNumberOfChanges(IRoundInformation currentRound,
            IRoundInformation nextRound);

       
        protected IList<int> GetCarrierTypeIds(IList<ICarrierInsertionRemovalFlagInformation> carriers)
        {
            return (from carrier in carriers
                    select carrier.CarrierTypeId)
                .ToList();
        }
    }
}
