﻿using System;
using System.Collections.Generic;

namespace PaintShopSchedulingCarrierChangeOptimizer.Solvers
{
    /// <summary>
    /// This class optimizes the carrier changes by reducing it to
    /// longest common subsequence problem. 
    /// Therefore a simple dynamic programming solution is used.
    /// </summary>
    internal class SimpleLongestCommonSubsequenceCarrierChangeOptimizer : AbstractCarrierChangeOptimizer
    {
        public override INumberOfCarrierChanges EvaluateNumberOfChanges(IRoundInformation currentRound, IRoundInformation nextRound)
        {
            // gather used carrier type ids of the current round
            IList<int> carrierTypesIdsInCurrentRound = GetCarrierTypeIds(currentRound.CarrierInsertionRemovalFlags);

            // gather used carrier type ids of the next round
            IList<int> carrierTypesIdsInNextRound = GetCarrierTypeIds(nextRound.CarrierInsertionRemovalFlags);

            int[,] longestCommonSubsequenceMatrix = GenerateLongestCommonSubsequenceMatrix(carrierTypesIdsInCurrentRound, carrierTypesIdsInNextRound);
            int numberOfUnforcedCarriersCurrentRound = carrierTypesIdsInCurrentRound.Count;
            int numberOfUnforcedCarriersNextRound = carrierTypesIdsInNextRound.Count;
            int longestCommonSubsequenceLength = longestCommonSubsequenceMatrix[numberOfUnforcedCarriersCurrentRound, numberOfUnforcedCarriersNextRound];
            int numberOfRemovals = numberOfUnforcedCarriersCurrentRound - longestCommonSubsequenceLength;
            int numberOfInsertions = numberOfUnforcedCarriersNextRound - longestCommonSubsequenceLength;

            return new NumberOfCarrierChanges(numberOfRemovals, numberOfInsertions);
        }

        protected override void Optimize(IList<ICarrierInsertionRemovalFlagInformation> preparedCurrentRoundCarrierList,
            IList<ICarrierInsertionRemovalFlagInformation> preparedNextRoundCarrierList)
        {
            // gather used carrier type ids of the current round
            IList<int> carrierTypesIdsInCurrentRound = GetCarrierTypeIds(preparedCurrentRoundCarrierList);

            // gather used carrier type ids of the next round
            IList<int> carrierTypesIdsInNextRound = GetCarrierTypeIds(preparedNextRoundCarrierList);

            int[,] longestCommonSubsequenceMatrix = GenerateLongestCommonSubsequenceMatrix(carrierTypesIdsInCurrentRound, carrierTypesIdsInNextRound);

            // These lists contain the indices of the elements of the longest common subsequence
            IList<int> currentRoundlongestCommonSubsequenceIndices = new List<int>();
            IList<int> nextRoundlongestCommonSubsequenceIndices = new List<int>();

            backtrackLongestCommonSubsequence(longestCommonSubsequenceMatrix,
                carrierTypesIdsInCurrentRound, carrierTypesIdsInNextRound,
                carrierTypesIdsInCurrentRound.Count, carrierTypesIdsInNextRound.Count,
                currentRoundlongestCommonSubsequenceIndices,
                nextRoundlongestCommonSubsequenceIndices);

            SetCarrierChangeFlags(preparedCurrentRoundCarrierList, preparedNextRoundCarrierList, 
                currentRoundlongestCommonSubsequenceIndices, nextRoundlongestCommonSubsequenceIndices);
        }

        /// <summary>
        /// This function returns a matrix. It contains for each index combination of the two given lists
        /// <paramref name="carrierTypesIdsInCurrentRound"/> and <paramref name="carrierTypesIdsInNextRound"/> 
        /// the length of the longest common subsequence.
        /// </summary>
        /// <param name="carrierTypesIdsInCurrentRound"></param>
        /// <param name="carrierTypesIdsInNextRound"></param>
        /// <returns>the dimensions are one more than the length of the rounds. 
        /// So to each index one must be added to get the desired result</returns>
        private int[,] GenerateLongestCommonSubsequenceMatrix(IList<int> carrierTypesIdsInCurrentRound,
            IList<int> carrierTypesIdsInNextRound)
        {
            // get amount of the used carrier type ids
            int currentRoundLength = carrierTypesIdsInCurrentRound.Count;
            int nextRoundLength = carrierTypesIdsInNextRound.Count;

            // This matrix is used to compute the length of the common longest subsequence
            // and for the gathering which elements of each list are in the common longest subsequence
            // Notice that the dimensions are one more than the length of the rounds. 
            // This is necessary so that comparison of the first element of each list has already a previous state.
            int[,] longestCommonSubsequenceMatrix = new int[currentRoundLength + 1, nextRoundLength + 1];

            // initialize the first row with 0
            for (int index = 0; index <= currentRoundLength; index++)
            {
                longestCommonSubsequenceMatrix[index, 0] = 0;
            }

            // initialize the first column with 0
            for (int index = 0; index <= nextRoundLength; index++)
            {
                longestCommonSubsequenceMatrix[0, index] = 0;
            }

            for (int currentRoundIndex = 1; currentRoundIndex <= currentRoundLength; currentRoundIndex++)
            {
                for (int nextRoundIndex = 1; nextRoundIndex <= nextRoundLength; nextRoundIndex++)
                {
                    // check if the values on the current positions in the two lists are equal
                    if (carrierTypesIdsInCurrentRound[currentRoundIndex - 1] == carrierTypesIdsInNextRound[nextRoundIndex - 1])
                    {
                        // if this is the case then the length of the longest common subsequence is the length of 
                        // the previous positions incremented by one
                        // --> set this value at the corresponding matrix position 
                        longestCommonSubsequenceMatrix[currentRoundIndex, nextRoundIndex] =
                            longestCommonSubsequenceMatrix[currentRoundIndex - 1, nextRoundIndex - 1] + 1;
                    }
                    else
                    {
                        // otherwise take the max value when one of the current elements is omitted for the subsequence
                        // --> set this value at the current matrix position 
                        longestCommonSubsequenceMatrix[currentRoundIndex, nextRoundIndex] =
                            Math.Max(
                                longestCommonSubsequenceMatrix[currentRoundIndex, nextRoundIndex - 1],
                                longestCommonSubsequenceMatrix[currentRoundIndex - 1, nextRoundIndex]);
                    }
                }
            }

            return longestCommonSubsequenceMatrix;
        }

        /// <summary>
        /// Adapts for the given rounds <paramref name="preparedCurrentRoundCarrierList"/> 
        /// and <paramref name="preparedNextRoundCarrierList"/> 
        /// the carrier removal respectively insertion flags.
        /// </summary>
        /// <param name="preparedCurrentRoundCarrierList"></param>
        /// <param name="preparedNextRoundCarrierList"></param>
        /// <param name="currentRoundlongestCommonSubsequenceIndices">set all removal flags from the <paramref name="preparedCurrentRoundCarrierList"/> 
        ///             which indices are not in this given list</param>
        /// <param name="nextRoundlongestCommonSubsequenceIndices">set all insertion flags from the <paramref name="preparedNextRoundCarrierList"/> 
        ///             which indices are not in this given list</param>
        private void SetCarrierChangeFlags(IList<ICarrierInsertionRemovalFlagInformation> preparedCurrentRoundCarrierList,
            IList<ICarrierInsertionRemovalFlagInformation> preparedNextRoundCarrierList, 
            IList<int> currentRoundlongestCommonSubsequenceIndices,
            IList<int> nextRoundlongestCommonSubsequenceIndices)
        {
            // Elements of the current round which are not in the longest common subsequence are not used in the next round.
            // Therefore set the removal flag for these carriers.
            for (int index = 0; index < preparedCurrentRoundCarrierList.Count; index++)
            {
                if (!currentRoundlongestCommonSubsequenceIndices.Contains(index))
                {
                    preparedCurrentRoundCarrierList[index].RemovalFlag = true;
                }
                else
                {
                    preparedCurrentRoundCarrierList[index].RemovalFlag = false;
                }
            }

            // Elements of the next round which are not in the longest common subsequence are not used in the current round.
            // Therefore set the insertion flag for these carriers.
            for (int index = 0; index < preparedNextRoundCarrierList.Count; index++)
            {
                if (!nextRoundlongestCommonSubsequenceIndices.Contains(index))
                {
                    preparedNextRoundCarrierList[index].InsertionFlag = true;
                }
                else
                {
                    preparedNextRoundCarrierList[index].InsertionFlag = false;
                }
            }
        }

        /// <summary>
        /// This function gathers the indices of the elements which are in the longest common subsequence. 
        /// The computed indices are stored in <paramref name="currentRoundlongestCommonSubsequenceIndices"/> and 
        /// <paramref name="nextRoundlongestCommonSubsequenceIndices"/>
        /// </summary>
        /// <param name="longestCommonSubsequenceMatrix"></param>
        /// <param name="currentRound">first list of elements</param>
        /// <param name="nextRound">second list of elements</param>
        /// <param name="currentRoundIndex">get indices of the longest common subsequence up to this given index position -1 of the current round</param>
        /// <param name="nextRoundIndex">get indices of the longest common subsequence up to this given index position -1 of the next round</param>
        /// <param name="currentRoundlongestCommonSubsequenceIndices">
        /// If this recursive function terminates then this list contains the indices of the elements of the common longest subsequence of the current round up to the given <paramref name="currentRoundIndex"/>.
        /// If the function is called from another function then this list must be an empty list.</param>
        /// <param name="nextRoundlongestCommonSubsequenceIndices">
        /// If this recursive function terminates then this list contains the indices of the elements of the common longest subsequence of the next round up to the given <paramref name="nextRoundIndex"/>.
        /// If the function is called from another function then this list must be an empty list.</param>
        private void backtrackLongestCommonSubsequence(
            int[,] longestCommonSubsequenceMatrix,
            IList<int> currentRound,
            IList<int> nextRound,
            int currentRoundIndex,
            int nextRoundIndex,
            IList<int> currentRoundlongestCommonSubsequenceIndices,
            IList<int> nextRoundlongestCommonSubsequenceIndices)
         {

            // stop the recursion if one of the currently considered indices is already 0.
            if (currentRoundIndex == 0 || nextRoundIndex == 0)
            {
                return;
            }

            // if the elements of the currently considered indices are equal
            if (currentRound[currentRoundIndex-1] == nextRound[nextRoundIndex-1])
            {
                // then add the index position to the corresponding list 
                currentRoundlongestCommonSubsequenceIndices.Add(currentRoundIndex-1);
                nextRoundlongestCommonSubsequenceIndices.Add(nextRoundIndex-1);

                // and call the function again where each index position is reduced by one. 
                backtrackLongestCommonSubsequence(
                    longestCommonSubsequenceMatrix,
                    currentRound, nextRound,
                    currentRoundIndex - 1, nextRoundIndex - 1,
                    currentRoundlongestCommonSubsequenceIndices,
                    nextRoundlongestCommonSubsequenceIndices);
                
            // otherwise, test which common subsequence is longer when one of the current elements is omitted
            } else if (longestCommonSubsequenceMatrix[currentRoundIndex, nextRoundIndex - 1] >
                longestCommonSubsequenceMatrix[currentRoundIndex - 1, nextRoundIndex])
            {

                // ommit the current element of the next round
                backtrackLongestCommonSubsequence(
                    longestCommonSubsequenceMatrix,
                    currentRound, nextRound,
                    currentRoundIndex, nextRoundIndex - 1,
                    currentRoundlongestCommonSubsequenceIndices,
                    nextRoundlongestCommonSubsequenceIndices);
            } else
            {
                // ommit the current element of the current round
                backtrackLongestCommonSubsequence(
                    longestCommonSubsequenceMatrix,
                    currentRound, nextRound,
                    currentRoundIndex - 1, nextRoundIndex,
                    currentRoundlongestCommonSubsequenceIndices,
                    nextRoundlongestCommonSubsequenceIndices);
            }

            
        }
    }
}
