﻿namespace PaintShopSchedulingSimpleTypes
{
    /// <summary>
    /// Can be used to perform calculations based on the positions in the schedule
    /// </summary>
    public static class PositionCalculator
    {
        /// <summary>
        /// Get the previous position coordinates for a given position in the schedule.
        /// The method assumes that the position is valid.
        /// </summary>
        /// <param name="position">Position index of the starting position.</param>
        /// <param name="schedule">The schedule that should be calculated with.</param>
        /// <returns>Position coordinates of the previous position, (-1,-1) if outside of the schedule.</returns>
        public static Position GetPreviousPosition(Position position, ISimpleSolution schedule)
        {
            return GetPreviousPosition(position.RoundIndex, position.PositionIndex, schedule);
        }

        /// <summary>
        /// Get the previous position coordinates for a given position in the schedule.
        /// The method assumes that the position is valid.
        /// </summary>
        /// <param name="roundIndex">Round index of the previous position (zero-indexed).</param>
        /// <param name="positionIndex">Position index of the previous position (zero-indexed).</param>
        /// <param name="schedule">The schedule that should be calculated with.</param>
        /// <returns>Position coordinates of the previous position, (-1,-1) if outside of the schedule.</returns>
        private static Position GetPreviousPosition(int roundIndex, int positionIndex, ISimpleSolution schedule)
        {
            if (positionIndex > 0)
            {
                // previous position is within same round
                return new Position(roundIndex, --positionIndex);
            }

            if (roundIndex <= 0)
            {
                // Position lies outside of schedule
                return new Position(-1, -1);
            }

            roundIndex--;

            while (roundIndex > -1 && schedule.GetNumberOfCarriersInRound(roundIndex) == 0)
            {
                roundIndex--;
            }

            if (roundIndex == -1) {
                return new Position(-1, -1);
            }

            // previous position is last carrier in previous round
            return new Position(roundIndex, schedule.GetNumberOfCarriersInRound(roundIndex) - 1);
        }
    }

}
