﻿using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class ForbiddenColorSequencesEvaluator : IForbiddenColorSequencesEvaluator
    {
        private int _totalViolations;
        private IList<int> _totalViolationsPerRound;
        private readonly SimpleInstance _instance;

        // colorid -> last appearance in number of positions from current position
        private IDictionary<int, int> _lastAppearancePositionOffset;

        private ICollection<SimpleForbiddenColorSequence> _forbiddenColorSequences;

        public ForbiddenColorSequencesEvaluator(SimpleInstance instance)
        {
            _instance = instance;
        }

        public void Evaluate(ICollection<SimpleForbiddenColorSequence> forbiddenColorSequences,
            ISimpleSolution solution)
        {
            _totalViolations = 0;
            _totalViolationsPerRound = Enumerable.Repeat(0, solution.NumberOfRounds - 1).ToList();
            _lastAppearancePositionOffset = new Dictionary<int, int>();
            _forbiddenColorSequences = forbiddenColorSequences;


            for (int roundIndex = 0; roundIndex < solution.NumberOfRounds; roundIndex++) {
                int numberOfCarriersInCurrentRound = solution.GetNumberOfCarriersInRound(roundIndex);

                for (int posIndex = 0; posIndex < numberOfCarriersInCurrentRound; posIndex++) {
                    int currentColorId =
                        solution.GetColorIdAtPosition(new Position(roundIndex, posIndex));

                    UpdateForbiddenColorSequenceData(currentColorId, roundIndex, posIndex);
                }
            }
        }

        private void UpdateForbiddenColorSequenceData(int currentColorId, int roundIndex, int posIndex)
        {
            if (roundIndex > 0) {
                ICollection<SimpleForbiddenColorSequence> forbiddenSequencesToCurrentColor =
                    GetForbiddenSequencesToCurrentColor(currentColorId);

                foreach (SimpleForbiddenColorSequence forbiddenColorSeq in forbiddenSequencesToCurrentColor) {
                    int fromColorId = forbiddenColorSeq.FromColor.Id;
                    if (_lastAppearancePositionOffset.ContainsKey(fromColorId) &&
                        _lastAppearancePositionOffset[fromColorId] <= forbiddenColorSeq.MinimumCarriersInBetween) {
                        _totalViolations++;
                        _totalViolationsPerRound[roundIndex - 1]++;

                        System.Console.WriteLine("Encountered forbidden color sequence in round " + roundIndex +
                                                 " at position " + posIndex + ": Color of type " +
                                                 _instance.ColorNames[currentColorId] +
                                                 " may not follow after type " +
                                                 _instance.CarrierTypeNames[fromColorId] +
                                                 " unless there are at least " +
                                                 forbiddenColorSeq.MinimumCarriersInBetween + " carriers in between.");
                    }
                }
            }

            // update current color in last appearance dictionary
            if (!_lastAppearancePositionOffset.ContainsKey(currentColorId)) {
                _lastAppearancePositionOffset.Add(currentColorId, 0);
            }
            else {
                _lastAppearancePositionOffset[currentColorId] = 0;
            }

            // increase last appearance counter for all saved appearances by one
            foreach (int colorId in _lastAppearancePositionOffset.Keys.ToList()) {
                _lastAppearancePositionOffset[colorId]++;
            }
        }

        private ICollection<SimpleForbiddenColorSequence> GetForbiddenSequencesToCurrentColor(int currentColorId)
        {
            return _forbiddenColorSequences.Where(x => x.ToColor.Id == currentColorId).ToList();
        }

        public int GetTotalViolations()
        {
            return _totalViolations;
        }

        public IList<int> GetViolationsPerRound()
        {
            return _totalViolationsPerRound;
        }
    }
}