﻿using System;
using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class UnfulfilledDemandEvaluator : IUnfulfilledDemandEvaluator
    {
        private int _unfulfilledDemands;
        private int _missingDemandPieces;
        private IList<int> _unfulfilledDemandsPerRound;
        private IList<int> _missingDemandPiecesPerRound;
        private ICollection<SimpleDemand> _demandsSortedByDueRound;
        private IDictionary<int, ICollection<SimpleMaterialCapacity>> _materialsForConfigIds;

        // round id -> material id -> color id -> available pieces
        private IList<IDictionary<int, IDictionary<int, int>>> _availablePiecesPerRound;

        public void Evaluate(ICollection<SimpleDemand> demands, ICollection<SimpleCarrierConfiguration> configurations,
            ISimpleSolution solution)
        {
            _demandsSortedByDueRound = demands.OrderBy(x => x.DueRound).ToList();
            _unfulfilledDemands = 0;
            _missingDemandPieces = 0;
            _unfulfilledDemandsPerRound = Enumerable.Repeat(0, solution.NumberOfRounds).ToList();
            _missingDemandPiecesPerRound = Enumerable.Repeat(0, solution.NumberOfRounds).ToList();

            InitializeMaterialsForConfigIds(configurations);
            InitializeAvailablePiecesPerRound(solution);

            foreach (SimpleDemand demand in _demandsSortedByDueRound) {
                UpdateViolationsForDemand(demand, solution);
            }
        }

        private void InitializeMaterialsForConfigIds(ICollection<SimpleCarrierConfiguration> configurations)
        {
            _materialsForConfigIds = new Dictionary<int, ICollection<SimpleMaterialCapacity>>();

            foreach (SimpleCarrierConfiguration config in configurations) {
                _materialsForConfigIds.Add(config.Id, config.MaterialCapacities);
            }
        }

        private void InitializeAvailablePiecesPerRound(ISimpleSolution solution)
        {
            _availablePiecesPerRound = new List<IDictionary<int, IDictionary<int, int>>>();

            for (int roundIndex = 1; roundIndex < solution.NumberOfRounds; roundIndex++) {
                IDictionary<int, IDictionary<int, int>> roundDictionary = new Dictionary<int, IDictionary<int, int>>();

                int numberOfCarriersInRound = solution.GetNumberOfCarriersInRound(roundIndex);
                for (int posIndex = 0; posIndex < numberOfCarriersInRound; posIndex++) {
                    SimpleCarrier carrierAtPos = solution.GetCarrierAtPosition(new Position(roundIndex, posIndex));
                    int configId = carrierAtPos.CarrierConfigurationId;
                    int colorId = carrierAtPos.ColorId;

                    if (configId <= -1) {
                        continue;
                    }

                    foreach (SimpleMaterialCapacity materialCapacity in _materialsForConfigIds[configId]) {
                        int materialId = materialCapacity.Material.Id;
                        int capacity = materialCapacity.Capacity;

                        if (!roundDictionary.ContainsKey(materialId)) {
                            roundDictionary.Add(materialId, new Dictionary<int, int>());
                        }

                        if (!roundDictionary[materialId].ContainsKey(colorId)) {
                            roundDictionary[materialId].Add(colorId, 0);
                        }

                        roundDictionary[materialId][colorId] += capacity;
                    }
                }

                _availablePiecesPerRound.Add(roundDictionary);
            }
        }

        private void UpdateViolationsForDemand(SimpleDemand demand, ISimpleSolution solution)
        {
            int materialId = demand.Material.Id;
            int colorId = demand.Color.Id;
            int missingPieces = demand.NumberOfPieces;
            int dueRound = demand.DueRound;

            for (int roundIndex = 0; roundIndex < solution.NumberOfRounds; roundIndex++) {
                if (roundIndex > 0 && _availablePiecesPerRound[roundIndex - 1].ContainsKey(materialId) &&
                    _availablePiecesPerRound[roundIndex - 1][materialId].ContainsKey(colorId)) {
                    int availablePiecesInRound = _availablePiecesPerRound[roundIndex - 1][materialId][colorId];
                    int consumedPieces = Math.Min(availablePiecesInRound, missingPieces);

                    missingPieces -= consumedPieces;
                    _availablePiecesPerRound[roundIndex - 1][materialId][colorId] -= consumedPieces;

                    if (missingPieces == 0) {
                        break;
                    }
                }

                if (roundIndex == dueRound && missingPieces > 0) {
                    _unfulfilledDemands++;
                    _missingDemandPieces += missingPieces;
                    Console.WriteLine("Demand " + demand.Id + " is unfulfilled: " + missingPieces +
                                      " pieces are missing or not scheduled before due round " + dueRound);

                    _unfulfilledDemandsPerRound[roundIndex]++;
                    _missingDemandPiecesPerRound[roundIndex] += missingPieces;
                }
            }
        }

        public int GetUnfulfilledDemands()
        {
            return _unfulfilledDemands;
        }

        public int GetMissingDemandPieces()
        {
            return _missingDemandPieces;
        }

        public IList<int> GetUnfulfilledDemandsPerRound()
        {
            return _unfulfilledDemandsPerRound;
        }

        public IList<int> GetMissingDemandPiecesPerRound()
        {
            return _missingDemandPiecesPerRound;
        }
    }
}